@@ -1,31 +1,43 @@
-""" Defines a KernelClient that provides thread-safe sockets with async callbacks on message replies.
+""" Defines a KernelClient that provides thread-safe sockets with async callbacks on message
+replies.
 """
-from __future__ import absolute_import
+import asyncio
 import atexit
-import errno
-import sys
-from threading import Thread, Event
 import time
+from concurrent.futures import Future
+from threading import Event, Thread
+from typing import Any, Dict, List, Optional
 
-# import ZMQError in top-level namespace, to avoid ugly attribute-error messages
-# during garbage collection of threads at exit:
-from zmq import ZMQError
-from zmq.eventloop import ioloop, zmqstream
+import zmq
+from tornado.ioloop import IOLoop
+from traitlets import Instance, Type
+from traitlets.log import get_logger
+from zmq.eventloop import zmqstream
+
+from .channels import HBChannel
+from .client import KernelClient
+from .session import Session
 
 # Local imports
-from traitlets import Type, Instance
-from jupyter_client.channels import HBChannel
-from jupyter_client import KernelClient
+# import ZMQError in top-level namespace, to avoid ugly attribute-error messages
+# during garbage collection of threads at exit
 
-class ThreadedZMQSocketChannel(object):
+
+class ThreadedZMQSocketChannel:
     """A ZMQ socket invoking a callback in the ioloop"""
+
     session = None
     socket = None
     ioloop = None
     stream = None
     _inspect = None
 
-    def __init__(self, socket, session, loop):
+    def __init__(
+        self,
+        socket: Optional[zmq.Socket],
+        session: Optional[Session],
+        loop: Optional[IOLoop],
+    ) -> None:
         """Create a channel.
 
         Parameters
@@ -35,9 +47,9 @@ class ThreadedZMQSocketChannel(object):
         session : :class:`session.Session`
             The session to use.
         loop
-            A pyzmq ioloop to connect the socket to using a ZMQStream
+            A tornado ioloop to connect the socket to using a ZMQStream
         """
-        super(ThreadedZMQSocketChannel, self).__init__()
+        super().__init__()
 
         self.socket = socket
         self.session = session
@@ -45,24 +57,54 @@ class ThreadedZMQSocketChannel(object):
         evt = Event()
 
         def setup_stream():
+            assert self.socket is not None
             self.stream = zmqstream.ZMQStream(self.socket, self.ioloop)
             self.stream.on_recv(self._handle_recv)
             evt.set()
 
+        assert self.ioloop is not None
         self.ioloop.add_callback(setup_stream)
         evt.wait()
 
     _is_alive = False
-    def is_alive(self):
+
+    def is_alive(self) -> bool:
+        """Whether the channel is alive."""
         return self._is_alive
 
-    def start(self):
+    def start(self) -> None:
+        """Start the channel."""
         self._is_alive = True
 
-    def stop(self):
+    def stop(self) -> None:
+        """Stop the channel."""
         self._is_alive = False
 
-    def close(self):
+    def close(self) -> None:
+        """Close the channel."""
+        if self.stream is not None and self.ioloop is not None:
+            # c.f.Future for threadsafe results
+            f: Future = Future()
+
+            def close_stream():
+                try:
+                    if self.stream is not None:
+                        self.stream.close(linger=0)
+                        self.stream = None
+                except Exception as e:
+                    f.set_exception(e)
+                else:
+                    f.set_result(None)
+
+            self.ioloop.add_callback(close_stream)
+            # wait for result
+            try:
+                f.result(timeout=5)
+            except Exception as e:
+                log = get_logger()
+                msg = f"Error closing stream {self.stream}: {e}"
+                log.warning(msg, RuntimeWarning, stacklevel=2)
+
         if self.socket is not None:
             try:
                 self.socket.close(linger=0)
@@ -70,7 +112,7 @@ class ThreadedZMQSocketChannel(object):
                 pass
             self.socket = None
 
-    def send(self, msg):
+    def send(self, msg: Dict[str, Any]) -> None:
         """Queue a message to be sent from the IOLoop's thread.
 
         Parameters
@@ -80,23 +122,29 @@ class ThreadedZMQSocketChannel(object):
         This is threadsafe, as it uses IOLoop.add_callback to give the loop's
         thread control of the action.
         """
+
         def thread_send():
+            assert self.session is not None
             self.session.send(self.stream, msg)
+
+        assert self.ioloop is not None
         self.ioloop.add_callback(thread_send)
 
-    def _handle_recv(self, msg):
+    def _handle_recv(self, msg_list: List) -> None:
         """Callback for stream.on_recv.
 
         Unpacks message, and calls handlers with it.
         """
-        ident,smsg = self.session.feed_identities(msg)
+        assert self.ioloop is not None
+        assert self.session is not None
+        ident, smsg = self.session.feed_identities(msg_list)
         msg = self.session.deserialize(smsg)
         # let client inspect messages
         if self._inspect:
             self._inspect(msg)
         self.call_handlers(msg)
 
-    def call_handlers(self, msg):
+    def call_handlers(self, msg: Dict[str, Any]) -> None:
         """This method is called in the ioloop thread when a message arrives.
 
         Subclasses should override this method to handle incoming messages.
@@ -106,14 +154,13 @@ class ThreadedZMQSocketChannel(object):
         """
         pass
 
-    def process_events(self):
+    def process_events(self) -> None:
         """Subclasses should override this with a method
         processing any pending GUI events.
         """
         pass
 
-
-    def flush(self, timeout=1.0):
+    def flush(self, timeout: float = 1.0) -> None:
         """Immediately processes all pending messages on this channel.
 
         This is only used for the IOPub channel.
@@ -133,37 +180,40 @@ class ThreadedZMQSocketChannel(object):
         # We do the IOLoop callback process twice to ensure that the IOLoop
         # gets to perform at least one full poll.
         stop_time = time.time() + timeout
-        for i in range(2):
+        assert self.ioloop is not None
+        for _ in range(2):
             self._flushed = False
             self.ioloop.add_callback(self._flush)
             while not self._flushed and time.time() < stop_time:
                 time.sleep(0.01)
 
-    def _flush(self):
+    def _flush(self) -> None:
         """Callback for :method:`self.flush`."""
+        assert self.stream is not None
         self.stream.flush()
         self._flushed = True
 
 
 class IOLoopThread(Thread):
-    """Run a pyzmq ioloop in a thread to send and receive messages
-    """
+    """Run a pyzmq ioloop in a thread to send and receive messages"""
+
     _exiting = False
     ioloop = None
 
     def __init__(self):
-        super(IOLoopThread, self).__init__()
+        """Initialize an io loop thread."""
+        super().__init__()
         self.daemon = True
 
     @staticmethod
     @atexit.register
-    def _notice_exit():
+    def _notice_exit() -> None:
         # Class definitions can be torn down during interpreter shutdown.
         # We only need to set _exiting flag if this hasn't happened.
         if IOLoopThread is not None:
             IOLoopThread._exiting = True
 
-    def start(self):
+    def start(self) -> None:
         """Start the IOLoop thread
 
         Don't return until self.ioloop is defined,
@@ -173,46 +223,38 @@ class IOLoopThread(Thread):
         Thread.start(self)
         self._start_event.wait()
 
-    def run(self):
+    def run(self) -> None:
         """Run my loop, ignoring EINTR events in the poller"""
-        if 'asyncio' in sys.modules:
-            # tornado may be using asyncio,
-            # ensure an eventloop exists for this thread
-            import asyncio
-            asyncio.set_event_loop(asyncio.new_event_loop())
-        self.ioloop = ioloop.IOLoop()
+        loop = asyncio.new_event_loop()
+        asyncio.set_event_loop(loop)
+        loop.run_until_complete(self._async_run())
+
+    async def _async_run(self):
+        self.ioloop = IOLoop.current()
         # signal that self.ioloop is defined
         self._start_event.set()
         while True:
-            try:
-                self.ioloop.start()
-            except ZMQError as e:
-                if e.errno == errno.EINTR:
-                    continue
-                else:
-                    raise
-            except Exception:
-                if self._exiting:
-                    break
-                else:
-                    raise
-            else:
+            await asyncio.sleep(1)
+            if self._exiting:
                 break
 
-    def stop(self):
+    def stop(self) -> None:
         """Stop the channel's event loop and join its thread.
 
         This calls :meth:`~threading.Thread.join` and returns when the thread
         terminates. :class:`RuntimeError` will be raised if
         :meth:`~threading.Thread.start` is called again.
         """
-        if self.ioloop is not None:
-            self.ioloop.add_callback(self.ioloop.stop)
+        self._exiting = True
         self.join()
         self.close()
         self.ioloop = None
 
-    def close(self):
+    def __del__(self):
+        self.close()
+
+    def close(self) -> None:
+        """Close the io loop thread."""
         if self.ioloop is not None:
             try:
                 self.ioloop.close(all_fds=True)
@@ -221,8 +263,7 @@ class IOLoopThread(Thread):
 
 
 class ThreadedKernelClient(KernelClient):
-    """ A KernelClient that provides thread-safe sockets with async callbacks on message replies.
-    """
+    """A KernelClient that provides thread-safe sockets with async callbacks on message replies."""
 
     @property
     def ioloop(self):
@@ -230,24 +271,32 @@ class ThreadedKernelClient(KernelClient):
 
     ioloop_thread = Instance(IOLoopThread, allow_none=True)
 
-    def start_channels(self, shell=True, iopub=True, stdin=True, hb=True):
+    def start_channels(
+        self,
+        shell: bool = True,
+        iopub: bool = True,
+        stdin: bool = True,
+        hb: bool = True,
+        control: bool = True,
+    ) -> None:
+        """Start the channels on the client."""
         self.ioloop_thread = IOLoopThread()
         self.ioloop_thread.start()
 
         if shell:
             self.shell_channel._inspect = self._check_kernel_info_reply
 
-        super(ThreadedKernelClient, self).start_channels(shell, iopub, stdin, hb)
+        super().start_channels(shell, iopub, stdin, hb, control)
 
-    def _check_kernel_info_reply(self, msg):
-        """This is run in the ioloop thread when the kernel info reply is received
-        """
-        if msg['msg_type'] == 'kernel_info_reply':
+    def _check_kernel_info_reply(self, msg: Dict[str, Any]) -> None:
+        """This is run in the ioloop thread when the kernel info reply is received"""
+        if msg["msg_type"] == "kernel_info_reply":
             self._handle_kernel_info_reply(msg)
             self.shell_channel._inspect = None
 
-    def stop_channels(self):
-        super(ThreadedKernelClient, self).stop_channels()
+    def stop_channels(self) -> None:
+        """Stop the channels on the client."""
+        super().stop_channels()
         if self.ioloop_thread.is_alive():
             self.ioloop_thread.stop()
 
@@ -255,3 +304,14 @@ class ThreadedKernelClient(KernelClient):
     shell_channel_class = Type(ThreadedZMQSocketChannel)
     stdin_channel_class = Type(ThreadedZMQSocketChannel)
     hb_channel_class = Type(HBChannel)
+    control_channel_class = Type(ThreadedZMQSocketChannel)
+
+    def is_alive(self) -> bool:
+        """Is the kernel process still running?"""
+        if self._hb_channel is not None:
+            # We don't have access to the KernelManager,
+            # so we use the heartbeat.
+            return self._hb_channel.is_beating()
+        # no heartbeat and not local, we can't tell if it's running,
+        # so naively return True
+        return True
